package weighting;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Random;
import java.util.Vector;

import structs.GameData;
import structs.MotionData;
import enumerate.Action;

/**
 * Stores weight for each possible action in a specific range
 * @author aldidoanta
 *
 */
public class ActionWeight implements Serializable{
	
	public ArrayList<ArrayList<ActionInfo>> weightList; //contains list of ActionInfo for each range segment
	public int[] range = {0, 100, 150}; //range segments
	
	//constructor
	public ActionWeight(GameData gameData, boolean playerNumber){
		
		//init Vector of motionData
		Vector<MotionData> motionData = gameData.getMyMotion(playerNumber);
		
		//init weightList ArrayList
		weightList = new ArrayList<ArrayList<ActionInfo>>();
		for (int i = 0; i < range.length; i++) {
			ArrayList<ActionInfo> actioninfoList = new ArrayList<ActionInfo>();
			
			//add all possible actionInfos for each range segment
			for (MotionData action : motionData) {
				int actionRange = calculateRange(action);
				if((actionRange >= range[i]) && (action.attackStartAddEnergy >= 0)){ //if the action meets the range requirement
					ActionInfo actioninfo = new ActionInfo(action.getMotionName(), motionData);
					actioninfoList.add(actioninfo);
				}
			}
			
			weightList.add(actioninfoList);
			
		}
		
	}
	
	/**
	 * calculates attack range of the action
	 * @param action
	 * @return attack range of the action
	 */
	public int calculateRange(MotionData action){
		
		int hitbox_left = action.getHit().getL();
		int hitbox_right = action.getHit().getR();
		int attackhitbox_right = action.getAttackHit().getR();
		int speedX = action.getSpeedX();
		int activeFrameNumber = action.getAttackActive();
		
		return Math.abs(attackhitbox_right - ((hitbox_left + hitbox_right) / 2))
				+ (speedX * activeFrameNumber);
	}
	
	/**
	 * determine which range index will be used
	 * @param rangeNumber
	 * @return selected range index
	 */
	public int selectRangeIndex(int rangeNumber){
		int maxRangeIdx = range.length - 1; //init with the highest range index (the longest)
		
		for(int i = 0; i < range.length; i++){
			if(rangeNumber < range[i]){
				maxRangeIdx = i-1;
			}
		}
		
		return maxRangeIdx;
	}

	/**
	 * Choose the best ActionInfo from the ArrayList of ActionInfo
	 */
	public ActionInfo chooseAction(Vector<MotionData> motionData, ArrayList<ActionInfo> actionChoice, String predictedActionName){
		ActionInfo action = actionChoice.get(0); //init with the first element of actionChoice;
		Random rand = new Random();
		
		if(predictedActionName == null){
			action = actionChoice.get(rand.nextInt(actionChoice.size()));
		}
		else{
			
			/* Iterate over possible action in the current range
			 * PRIORITY: 1. weight
			 * 			 2. number of frames
			 * 			 3. damage
			*/
			for (int i = 1; i < actionChoice.size(); i++) {
				ActionInfo actionCandidate = actionChoice.get(i); 
				 if(actionCandidate.weight >= action.weight){ //compare weight
					 if(actionCandidate.weight > action.weight){
						 action = actionCandidate; //update choosen action
					 }
					 else if(actionCandidate.weight == action.weight){
						 if(actionCandidate.frameNumber <= action.frameNumber){ //compare number of frames
							 if((actionCandidate.frameNumber < action.frameNumber)
							 && (actionCandidate.frameNumber < motionData.elementAt(Action.valueOf(predictedActionName).ordinal()).getFrameNumber())){
								 action = actionCandidate; //update choosen action
							 }
							 else if(actionCandidate.frameNumber == action.frameNumber){
								 if(actionCandidate.damage >= action.damage){ //compare damage
									 if((actionCandidate.damage > action.damage)
									 && (actionCandidate.damage > motionData.elementAt(Action.valueOf(predictedActionName).ordinal()).getAttackHitDamage())){
										 action = actionCandidate; //update choosen action
									 }
									 else{ // actionCandidate.damage == action.damage
										 //decide between the two ActionInfo randomly
										 action = rand.nextInt(2) == 0 ? action : actionCandidate;
									 }
								 }
							 }
						 }
					 }
				 }
			}	
		}
		
		return action;
	}
	
}
