import enumerate.Action;
import gameInterface.AIInterface;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;

import learner.J48Learner;
import opponentData.opponentData;
import structs.FrameData;
import structs.GameData;
import structs.Key;
import weighting.ActionInfo;
import weighting.ActionWeight;
import weka.core.DenseInstance;
import weka.core.Instance;

import commandcenter.CommandCenter;

/**
 * AI with OM approach using J48
 * @author aldidoanta
 *
 */
public class FICE_AI_OM implements AIInterface {
		
	//default attributes
	Key inputKey;
	boolean playerNumber;
	GameData gameData;
	FrameData frameData;
	CommandCenter cc;
	
	//attributes used to record opponent data
	File file, actionweightFile;
	PrintWriter writer;
	BufferedReader reader;
	ArrayList<opponentData> opponentRecord; //this ArrayList will be written in the log file
	boolean isOppStartAction; //a flag to check if the opponent has started a new action
	Key oppKey; //the key used by the opponent
	Key prev_oppKey; //the previous key used by the opponent
	String[] noiseAction = {"AIR_GUARD_RECOV",
							"AIR_RECOV",
							"CROUCH_GUARD_RECOV",
							"CROUCH_RECOV",
							"STAND_GUARD_RECOV",
							"STAND_RECOV",
							"THROW_HIT",
							"THROW_SUFFER"}; //these actions will not be recorded
	
	//used for learning
	J48Learner j48Learner;
	//used for action weighting
	ArrayList<ActionInfo> actionChoice;
	ActionInfo chosenAction;
	ActionWeight actionweight;
	int myHPBefore;
	int myHPAfter;
	int oppHPBefore;
	int oppHPAfter;
	long currentRemainingTime;
	
	//constants
	final String LOG_FILENAME = "data/aiData/FICE_AI_OM/logOM.arff"; 	//constant for file name
	final int NUM_ATTRIBUTE = 15;//constant for number of attributes in training data (including class attribute) 
	final int UPDATE_FREQUENCY = 6000; //update frequency in milliseconds
	final String[] STATES = {"STAND", "CROUCH", "AIR", "DOWN"};
	final String[] ACTIONS = {"AIR", "AIR_GUARD", "BACK_JUMP", "BACK_STEP", "CHANGE_DOWN", "CROUCH", "CROUCH_GUARD", "DASH", "DOWN", "FOR_JUMP", "FORWARD_WALK", "JUMP", "LANDING", "NEUTRAL", "RISE", "STAND", "STAND_GUARD", "STAND_F_D_DFA", "STAND_D_DF_FC", "STAND_A", "STAND_B", "STAND_D_DF_FA", "STAND_D_DF_FB", "STAND_FA", "STAND_D_DB_BB", "AIR_FA", "AIR_D_DB_BA", "STAND_F_D_DFB", "AIR_D_DF_FB", "AIR_B", "AIR_D_DF_FA", "AIR_F_D_DFB", "AIR_A", "AIR_D_DB_BB", "AIR_DA", "AIR_DB", "AIR_F_D_DFA", "AIR_FB", "AIR_UB", "AIR_UA", "STAND_FB", "STAND_D_DB_BA", "CROUCH_A", "CROUCH_FA", "CROUCH_B", "CROUCH_FB", "THROW_A", "THROW_B"};
	
	@Override
	public void close() {
		
	}

	@Override
	public String getCharacter() {
		return CHARACTER_ZEN;
	}

	@Override
	public void getInformation(FrameData _frameData) {
		frameData = _frameData;
		cc.setFrameData(frameData, playerNumber);
		
		//check training data update condition
		if(frameData.getRemainingTime() < 0){
			currentRemainingTime = 0;
		}
		if((currentRemainingTime == 0) && (frameData.getRemainingTime() > 0)){
			currentRemainingTime = frameData.getRemainingTime(); 
		}
		if(currentRemainingTime - frameData.getRemainingTime() >= UPDATE_FREQUENCY)
		{
			currentRemainingTime = frameData.getRemainingTime();
			
			try {
				writeToLogFile(); //write opponent data to log file
				j48Learner.buildClassifier();
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

	}

	@Override
	public int initialize(GameData _gameData, boolean _playerNumber) {
		
		//game and agent init
		this.inputKey = new Key();
		playerNumber = _playerNumber;
		gameData = _gameData;
		frameData = new FrameData();
		cc = new CommandCenter();
		
		//training data and action weight init
		createTrainData();
		opponentRecord = new ArrayList<opponentData>();

		//learner init
		try {
			j48Learner = new J48Learner();
			j48Learner.openFile(LOG_FILENAME);
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		//opponent action tracker init
		isOppStartAction = false;
		oppKey = new Key();
		prev_oppKey = new Key();

		
		//Weighting system init
		chosenAction = null;
		myHPBefore = 0;
		myHPAfter = 0;
		oppHPBefore = 0;
		oppHPAfter = 0;
		currentRemainingTime = 0;
		
		//init actionweight
		if (actionweight == null){
			//init actionweight if actionweight file is empty
			actionweight = new ActionWeight(_gameData, _playerNumber);
		}
		
		return 0;
	}

	@Override
	public Key input() {
		return inputKey;
	}
	
	@Override
	public void processing() {
		if (!frameData.getEmptyFlag() && frameData.getRemainingTime() > 0) {
			
			//get HP value after executing an action (if any)
			myHPAfter = cc.getMyHP();
			oppHPAfter = cc.getEnemyHP();
			//calculate reward given by the action
			if(chosenAction != null){
				int reward = calculateReward(chosenAction.damage);
				//update weight
				
				if(reward == 0){
					chosenAction.weight--;
				}
				else{
					chosenAction.weight += reward;	
				}
				
				updateWeight(actionChoice, chosenAction);
			}
			
			if (cc.getskillFlag()) { //if a command is being input
				inputKey = cc.getSkillKey();
			}
			else {
				inputKey.empty();
				cc.skillCancel();
					
				if(isKeyEmpty(prev_oppKey)){ //get opponent key for a new action
					oppKey = frameData.getKeyData().getOpponentKey(playerNumber);	
				}
				
				//get opponent action information
				if((!isOppStartAction) && (!isKeyEmpty(oppKey))){
					
					getOpponentInformation(true); //record opponent information
					isOppStartAction = true;
				}
				else if((isOppStartAction) && (!isKeyEmpty(oppKey))){
					isOppStartAction = false;
					oppKey.empty();
				}
				
				/* agent decision making */
				try {
					//get current situation
					opponentData opponentdata = getOpponentInformation(false);
					
					//check conditions for default rules
					int defaultRuleNumber = isDefaultSatisfied(opponentdata); 
					if(defaultRuleNumber != -1){
						chosenAction = null; //because we don't want to update the reward from default actions
						doDefaultAction(defaultRuleNumber);
					}
					else { //if default rules are not satisfied
						
						if(j48Learner.instances.numInstances() == 0){ //if training data is empty 
							actionChoice = actionweight.weightList.get(actionweight.selectRangeIndex(opponentdata.x));
							chosenAction = actionweight.chooseAction(gameData.getMyMotion(playerNumber), actionChoice, null);
							cc.commandCall(chosenAction.actionName); //execute action
						}
						else{
							//convert opponentData to Weka's instance
							Instance newInstance = opponentDataToInstance(opponentdata);
							
							//enemy's action prediction
							String actionPrediction = j48Learner.classifyInstance(newInstance, ACTIONS);
							
							//agent's action selection 
							actionChoice = actionweight.weightList.get(actionweight.selectRangeIndex(opponentdata.x));
							chosenAction = actionweight.chooseAction(gameData.getMyMotion(playerNumber), actionChoice, actionPrediction);
							
							//get HP value before executing an action
							myHPBefore = cc.getMyHP();
							oppHPBefore = cc.getEnemyHP();
							cc.commandCall(chosenAction.actionName); //execute action
						}
					}
					
				} catch (Exception e) {
					e.printStackTrace();
				}
				
				prev_oppKey = frameData.getKeyData().getOpponentKey(playerNumber); //record the last key used by the opponent
				
			}
		}
	}
	
	/* additional methods */
	/**
	 * Gets the horizontal position of opponent's current attack
	 * @param	enemyAction	the current Action of the enemy
	 * @return	X position of the opponent's current attack 
	 */
	public int getAttackRange_X(Action enemyAction){
		
		int result = 0;
		int enemy_atthitbox_R = gameData.getOpponentMotion(playerNumber).elementAt(enemyAction.ordinal()).getAttackHit().getR();
		int enemy_atthitbox_L = gameData.getOpponentMotion(playerNumber).elementAt(enemyAction.ordinal()).getAttackHit().getL();
		if(cc.getEnemyCharacter().isFront()){ //checks enemy facing direction
			int enemy_hitbox_R = cc.getEnemyCharacter().getRight();
			result = enemy_hitbox_R + (enemy_atthitbox_R - enemy_atthitbox_L);
		}
		else{
			int enemy_hitbox_L = cc.getEnemyCharacter().getLeft();
			result = enemy_hitbox_L - (enemy_atthitbox_R - enemy_atthitbox_L);
		} 

		return result;
	}
	
	/**
	 * Checks if the position of the character is within the opponent's attack range
	 * @param	enemyAction	the current Action of the enemy
	 * @return	if the hitbox of the character collides with the predicted opponent's attack hitbox
	 */
	public boolean isInEnemyAttRange(Action enemyAction){
		
		boolean result = false;
		if(cc.getMyCharacter().isFront()){ 	//checks character facing direction
			result = cc.getMyCharacter().getRight() > this.getAttackRange_X(enemyAction);
		}
		else{
			result = cc.getMyCharacter().getLeft() < this.getAttackRange_X(enemyAction);
		}
		return result;
	}
	
	
	
	/**
	 * Gets opponent information as (X, Y, Attack, Count)<p>
	 */
	public opponentData getOpponentInformation(boolean isIntanceRecorded){
		
		structs.CharacterData oppCharData = cc.getEnemyCharacter();
		structs.CharacterData myCharData = cc.getMyCharacter();
		
		int x = cc.getDistanceX();
		int y = cc.getDistanceY();
		int oppSpeedX = oppCharData.getSpeedX();
		int oppSpeedY = oppCharData.getSpeedY();
		int mySpeedX = myCharData.getSpeedX();
		int mySpeedY = myCharData.getSpeedX();
		String oppState = oppCharData.getState().name();
		String myState = myCharData.getState().name();
		int oppRemainingFrame = oppCharData.getRemainingFrame();
		int myRemainingFrame = myCharData.getRemainingFrame();
		int oppHP = cc.getEnemyHP();
		int myHP = cc.getMyHP();
		int oppEnergy = cc.getEnemyEnergy();
		int myEnergy = cc.getMyEnergy();
		String oppAction = oppCharData.getAction().name();
		
		//create a new opponentData
		opponentData newData = new opponentData(x,
				y,
				oppSpeedX,
				oppSpeedY,
				mySpeedX,
				mySpeedY,
				oppState,
				myState,
				oppRemainingFrame,
				myRemainingFrame,
				oppHP,
				myHP,
				oppEnergy,
				myEnergy,
				oppAction);
		
		if(isIntanceRecorded){ //if current information want to be added as training data
			
			if(!oppState.equals("DOWN")
			   && !Arrays.asList(noiseAction).contains(oppAction)){ //we don't want to record action while the opponent is down, or if the action is in the noiseAction array
			  
			   opponentRecord.add(newData); //create a new record
			}
		}
			
		return newData;		
	}
	
	/**
	 * Converts opponentData to Instance 
	 */
	public Instance opponentDataToInstance(opponentData od){
		//create instance as return value
		Instance newInstance = new DenseInstance(NUM_ATTRIBUTE);
		newInstance.setDataset(j48Learner.instances);
		
		newInstance.setValue(0, od.x);
		newInstance.setValue(1, od.y);
		newInstance.setValue(2, od.oppSpeedX);
		newInstance.setValue(3, od.oppSpeedY);
		newInstance.setValue(4, od.mySpeedX);
		newInstance.setValue(5, od.mySpeedY);
		newInstance.setValue(6, Arrays.asList(STATES).indexOf(od.oppState));
		newInstance.setValue(7, Arrays.asList(STATES).indexOf(od.myState));
		newInstance.setValue(8, od.oppRemainingFrame);
		newInstance.setValue(9, od.myRemainingFrame);
		newInstance.setValue(10, od.oppHP);
		newInstance.setValue(11, od.myHP);
		newInstance.setValue(12, od.oppEnergy);
		newInstance.setValue(13, od.myEnergy);
		newInstance.setValue(14, Arrays.asList(ACTIONS).indexOf(od.oppAction));
		
		return newInstance;
	}
	
	/**
	 * Calculates reward of previous action <p>
	 * The function calculates agent's HP difference before and after action, compared with opponent's HP difference
	 * @return reward value <p>
	 * 		   reward > 0, the action is more effective than the opponent's, vice versa
	 */
	public int calculateReward(int actionDamage){
		//TODO experimental reward calculation
		int result = (myHPAfter - myHPBefore) - (oppHPAfter - oppHPBefore);
		
		return result;
	}
	
	public void updateWeight(ArrayList<ActionInfo> actionList, ActionInfo chosenAct){
		
		for (ActionInfo actionInfo : actionList) {
			if(actionInfo.actionName.equals(chosenAct.actionName)){
				actionList.set(actionList.indexOf(actionInfo), chosenAct);
				break;
			}
		}
	}
	
	/**
	 * Creates a new ARFF file with its attribute definition
	 */
	public void createTrainData(){
		file = new File(LOG_FILENAME);
		
		try {
			writer = new PrintWriter(new BufferedWriter(new FileWriter(file,false)));
			
			//write ARFF file header
			writer.println("@relation enemymove");
			writer.println();
			writer.println("@attribute x numeric");
			writer.println("@attribute y numeric");
			writer.println("@attribute oppSpeedX numeric");
			writer.println("@attribute oppSpeedY numeric");
			writer.println("@attribute mySpeedX numeric");
			writer.println("@attribute mySpeedY numeric");
			writer.println("@attribute oppState {STAND, CROUCH, AIR, DOWN}");
			writer.println("@attribute myState {STAND, CROUCH, AIR, DOWN}");
			writer.println("@attribute oppRemainingFrame numeric");
			writer.println("@attribute myRemainingFrame numeric");
			writer.println("@attribute oppHP numeric");
			writer.println("@attribute myHP numeric");
			writer.println("@attribute oppEnergy numeric");
			writer.println("@attribute myEnergy numeric");
			writer.println("@attribute oppAction {AIR, AIR_GUARD, BACK_JUMP, BACK_STEP, CHANGE_DOWN, CROUCH, CROUCH_GUARD, DASH, DOWN, FOR_JUMP, FORWARD_WALK, JUMP, LANDING, NEUTRAL, RISE, STAND, STAND_GUARD, STAND_F_D_DFA, STAND_D_DF_FC, STAND_A, STAND_B, STAND_D_DF_FA, STAND_D_DF_FB, STAND_FA, STAND_D_DB_BB, AIR_FA, AIR_D_DB_BA, STAND_F_D_DFB, AIR_D_DF_FB, AIR_B, AIR_D_DF_FA, AIR_F_D_DFB, AIR_A, AIR_D_DB_BB, AIR_DA, AIR_DB, AIR_F_D_DFA, AIR_FB, AIR_UB, AIR_UA, STAND_FB, STAND_D_DB_BA, CROUCH_A, CROUCH_FA, CROUCH_B, CROUCH_FB, THROW_A, THROW_B}");
			writer.println();
			writer.println("@data");
			writer.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Writes the opponent data to a log file
	 */
	public void writeToLogFile(){
		try {
			writer = new PrintWriter(new BufferedWriter(new FileWriter(file,true)));
			for (opponentData data : opponentRecord)
			{
				//write data to file
				writer.println(data.x+","
								+data.y+","
								+data.oppSpeedX+","
								+data.oppSpeedY+","
								+data.mySpeedX+","
								+data.mySpeedY+","
								+data.oppState+","
								+data.myState+","
								+data.oppRemainingFrame+","
								+data.myRemainingFrame+","
								+data.oppHP+","
								+data.myHP+","
								+data.oppEnergy+","
								+data.myEnergy+","
								+data.oppAction
								);
			}
			writer.close();
			
			//clear opponentRecord
			opponentRecord.clear();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Checks if a Key is empty (no button pressed)
	 * @param K the key
	 * @return true if K is empty
	 */
	public boolean isKeyEmpty(Key K){
		return ((K.getLever() == 5) && !(K.A) && !(K.B) && !(K.C));
	}
	
	/**
	 * checks if current situation satisfies one of the conditions for default actions
	 * @return condition code number, -1 if there are no default conditions satisfied
	 */
	public int isDefaultSatisfied(opponentData od){
		int result = -1;
		
		//energy usage
		if ((od.myEnergy >= 300) && (od.x >= 300) && (od.oppState.equals("STAND"))){
			result = 1;	
		}
		else if((od.myEnergy >= 20) && (od.x <= 200) && (od.oppState.equals("STAND"))){
			result = 2;
		}
		else if((od.myEnergy >= 50) && (od.x <= 250)){
			result = 3;
		}
		else if((od.myEnergy >= 50) && (od.x <= 300) && (!od.oppState.equals("AIR"))){
			result = 4;
		}
		
		return result;
	}
	
	/**
	 * gives command to bot in correspondence with the satisfied default rule <p>
	 * see isDefaultSatisfied() for a list of default conditions
	 */
	public void doDefaultAction(int ruleNumber){
		
		switch(ruleNumber){
			case 1:{
				cc.commandCall("STAND_D_DF_FC");
				break;
			}
			case 2:{
				cc.commandCall("THROW_B");
				break;
			}
			case 3:{
				cc.commandCall("STAND_F_D_DFB");
				break;
			}
			case 4:{
				cc.commandCall("STAND_D_DB_BB");
				break;
			}
			
		}
	}
}
