package StrategiesClass;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.Deque;
import java.util.LinkedList;
import java.util.Random;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import commandcenter.CommandCenter;
import enumerate.Action;
import enumerate.State;
import simulator.Simulator;
import structs.FrameData;
import structs.GameData;
import structs.Key;

public class StrategySelecter {

	private File strFile;
	private BufferedReader bufReader;
	private boolean playerNumber;
	private Random rnd;

	public InitiativeStrategies AIStrategies;
	private PredictNextSkill preNextSkill;
	private MoveStrategies moveStr;
	private Simulator simulator;
	private Deque<Action> myAct;
	private Deque<Action> oppAct;

	private String[][] NPASkill;

	private String bestSkill = null;

	private boolean forceGuard;
	private int GuardTime = 0;
	
	public StrategySelecter(boolean thePlayerNumber, GameData gameData) {

		playerNumber = thePlayerNumber;
		strFile = new File("data/aiData/AI_ZBY0323/AISTR.txt");
		NPASkill = new String[26][];

		readNPASkill();

		rnd = new Random();

		AIStrategies = new InitiativeStrategies(playerNumber, NPASkill);
		simulator = gameData.getSimulator();
		moveStr = new MoveStrategies();
		forceGuard = false;
		preNextSkill = new PredictNextSkill(gameData, playerNumber);

		this.myAct = new LinkedList<Action>();
		this.oppAct = new LinkedList<Action>();
	}

	public void readNPASkill() {
		Pattern pattern = Pattern.compile("NPAS,.*");
		String strLine = "";
		int readLine = 0;

		try {
			bufReader = new BufferedReader(new FileReader(strFile));
			while ((strLine = bufReader.readLine()) != null) {
				Matcher matcher = pattern.matcher(strLine);
				if (matcher.find()) {
					strLine = matcher.group();
					NPASkill[readLine] = strLine.replaceAll("NPAS,", "").split(",");
					readLine++;
				}
			}
			bufReader.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	public void selectNextSkill(FrameData frameData, CommandCenter atkcc, Key inputKey) {
		Action myAct = frameData.getMyCharacter(playerNumber).getAction();
		Action oppAct = frameData.getOpponentCharacter(playerNumber).getAction();
		int oppSkillNum = -1;

		for (int i = 0; i < NPASkill.length; i++) {
			if (NPASkill[i][0].equals(oppAct.name())) {
				oppSkillNum = i;
			}
		}
		int oppSkillStart = 0;
		int oppSkillAct = 0;
		int oppSkillRec = 0;
		int oppremainingFrame = frameData.getOpponentCharacter(playerNumber).getRemainingFrame();

		if (oppSkillNum != -1) {
			oppSkillStart = Integer.parseInt(NPASkill[oppSkillNum][3]);
			oppSkillAct = Integer.parseInt(NPASkill[oppSkillNum][4]);
			oppSkillRec = Integer.parseInt(NPASkill[oppSkillNum][5]);
		}

		if (!forceGuard) {
			FrameData simNextS = simulateNextSkill(frameData, myAct, oppAct, oppSkillStart + oppSkillAct);
			// If the opponent can hit me,choose the guard type and use it.
			if (simNextS.getMyCharacter(playerNumber).getHp() < frameData.getMyCharacter(playerNumber).getHp()) {
				if (oppSkillNum != -1) {
					for (int n = 0; n < NPASkill.length; n++) {
						if ((Integer.parseInt(NPASkill[n][3]) + 1) < (oppSkillStart + oppSkillAct)) {
							bestSkill = NPASkill[n][0];
						}
					}
				} else {
					String tempAct = AIStrategies.selectNextSkill(frameData);
					if (tempAct != null) {
						bestSkill = tempAct;
					} else {
						Action bestSkillAction = preNextSkill.selectNextSkill(frameData);
						if (bestSkillAction != null){
							bestSkill = bestSkillAction.name();
						}
						else
							bestSkill = moveStr.selectMoveStr();
					}
				}
				if (bestSkill != null) {
					atkcc.commandCall(bestSkill);
				} else {
					chooseGuard(NPASkill[oppSkillNum][2], frameData, inputKey);
					GuardTime = oppSkillStart + oppSkillAct + oppSkillRec;
				}
			}
			// Else choose a skill which can hit the opponent and use it.
			else {
				if (oppSkillNum != -1) {
					int bestDamage = 0;
					for (int j = 0; j < NPASkill.length; j++) {
						simNextS = simulateNextSkill(frameData, Action.valueOf(NPASkill[j][0]), oppAct,
								oppremainingFrame);
						int tempDamage = frameData.getOpponentCharacter(playerNumber).getHp()
								- simNextS.getOpponentCharacter(playerNumber).getHp();
						if (tempDamage > bestDamage) {
							bestDamage = tempDamage;
							bestSkill = NPASkill[j][0];
						}
					}
				} else {
					String tempAct = AIStrategies.selectNextSkill(frameData);
					if (tempAct != null) {
						bestSkill = tempAct;
					} else {
						Action bestSkillAction = preNextSkill.selectNextSkill(frameData);
						if (bestSkillAction != null){
							bestSkill = bestSkillAction.name();
						}
						else
							bestSkill = moveStr.selectMoveStr();
					}
				}
				if (bestSkill != null) {
					atkcc.commandCall(bestSkill);
				}
			}
		} else {
			GuardTime--;
			if (GuardTime <= 0) {
				GuardTime = 0;
				cancelGuard(inputKey);
			}

		}
		bestSkill = null;
	}

	public void chooseGuard(String atkType, FrameData frameData, Key inputKey) {
		boolean isAir = frameData.getMyCharacter(playerNumber).getState().equals(State.AIR);
		if (isAir) {
			guard(1, frameData, inputKey);
		} else {
			switch (atkType) {
			case "H":
				guard(rnd.nextInt(1) + 1, frameData, inputKey);
				break;
			case "M":
				guard(1, frameData, inputKey);
				break;
			case "L":
				guard(2, frameData, inputKey);
				break;
			default:
			}
		}
	}

	public void guard(int GuardType, FrameData frameData, Key inputKey) {
		switch (GuardType) {
		case 1:
			standGuard(frameData, inputKey);
			break;
		case 2:
			crouchGuard(frameData, inputKey);
			break;
		case 3:
			airGuard(frameData, inputKey);
			break;
		default:
			crouchGuard(frameData, inputKey);
		}
	}

	public void standGuard(FrameData frameData, Key inputKey) {
		forceGuard = true;
		inputKey.empty();
		if (frameData.getMyCharacter(playerNumber).isFront())
			inputKey.L = true;
		else
			inputKey.R = true;
	}

	public void crouchGuard(FrameData frameData, Key inputKey) {
		forceGuard = true;
		inputKey.empty();
		inputKey.D = true;
		if (frameData.getMyCharacter(playerNumber).isFront())
			inputKey.L = true;
		else
			inputKey.R = true;
	}

	public void airGuard(FrameData frameData, Key inputKey) {
		forceGuard = true;
		inputKey.empty();
		if (frameData.getMyCharacter(playerNumber).isFront())
			inputKey.L = true;
		else
			inputKey.R = true;
	}

	public void updataGuard(FrameData frameData, Key inputKey) {
		if (forceGuard) {
			if (frameData.getMyCharacter(playerNumber).isFront()) {
				inputKey.L = true;
				inputKey.R = false;
			} else {
				inputKey.L = false;
				inputKey.R = true;
			}
		}
	}

	public void cancelGuard(Key inputKey) {
		forceGuard = false;
		inputKey.empty();
	}

	public FrameData simulateNextSkill(FrameData frameData, Action myAction, Action oppAction, int delay) {

		myAct.clear();
		myAct.add(myAction);

		oppAct.clear();
		oppAct.add(oppAction);

		FrameData simulatedFrame;

		simulatedFrame = simulator.simulate(frameData, playerNumber, myAct, oppAct, delay);

		simulatedFrame.getMyCharacter(playerNumber).getHp();
		simulatedFrame.getOpponentCharacter(playerNumber).getHp();

		return simulatedFrame;

	}

	public void printNPASkill() {

		for (int i = 0; i < NPASkill.length; i++) {
			for (int j = 0; j < NPASkill[i].length; j++) {
				System.out.print(NPASkill[i][j]);
				if (j != NPASkill[i].length - 1)
					System.out.print(",");
			}
			System.out.println();
		}

	}

}
